/**  FlexPainter Beta version 0.40
 * # ToDo # Totally free with credit is fine
 *
 * Copyright (c) 2019 by CVH.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and/or distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice appear
 * in all copies.
 *
 * With the unmatched assistance of Andrew Mustun.
 * This script is donated to the QCAD project.
 * Copyright (c) 2011-2019 by Andrew Mustun. All rights reserved.
 *
 * QCAD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * QCAD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with QCAD.
 */

include("../MathExamples.js");
include("scripts/simple.js");    // Added for addSpline simple API
//# Issue Solved # Use "scripts/simple.js" istead of "simple.js"
//https://qcad.org/rsforum/viewtopic.php?t=6528


/**
 * \class FlexPainter
 * \ingroup ecma_misc_examples_mathexamples
 * \brief Cast and morphs flexible painters over selected based entities.
 *
 */
function FlexPainter(guiAction) {
    MathExamples.call(this, guiAction);
};

FlexPainter.prototype = new MathExamples();
FlexPainter.includeBasePath = includeBasePath;


/**
 * This action paints over all selected line based entities from start to end and minimalist morphs the painter to the curvature
 * Outputs common fitpoint splines the user can edit with any tool available
 * Results may be better with radii of curvature of about the painters actual width or larger
 */
FlexPainter.prototype.beginEvent = function() {
    MathExamples.prototype.beginEvent.call(this);

//debugger;

    // FlexPainter.init states "Requires Document"
    // A document should be available....Retreive document, units
    var document = this.getDocument();
    this.units = document.getUnit();
    var di = this.getDocumentInterface();
    var ids = document.querySelectedEntities();    // Query Qlist with selected
    const idn = ids.length;    // Number of selected entities

    // Terminate without document units
    if (!isNumber(this.units) || isNull(this.units)){
        // Push warning to command history and terminate
        qDebug("NO Doc units");
        EAction.handleUserWarning("NO Doc units");
        EAction.handleUserMessage("Nothing casted, command terminated");
        debugger;    // Can't handle no units. Halt in debugger mode
        this.terminate();
        return;
    }

    // Push command start to command history
    EAction.handleUserCommand("MathExamples: Cast FlexPainter");
    // Terminate without any entity selected
    if (idn == 0) {
        // Push warning to command history and terminate
        EAction.handleUserInfo("Please, make a selection!");
        EAction.handleUserMessage("Nothing casted, command terminated");
        this.terminate();
        return;
    }

    // Show Dialog & terminate if cancelled
    if(!this.showDialog()) {
        // Push cancelled to command history and terminate
        EAction.handleUserMessage("Cancelled, command terminated");
        this.terminate();
        return;
    }

//    EAction.activateMainWindow();  // Ain't necessary
//    disableInput();  // Ain't necessary
//debugger;

    // Cronometration
    var crono = Date.now();

    // Retrieve painters specifics
    var painterLength = this.mainData[0][2];
    var painterStepSize = this.mainData[0][3];
    var knotlength = this.knotData[0][2];
    var appendiceslength = 0.00;
    // Sum used appendice lengths, to do is existing
    if (this.doBeginApex) appendiceslength += this.beginApexData[0][2];
    if (this.doEndApex) appendiceslength += this.endApexData[0][2];
    if (this.doInserts) appendiceslength += knotlength * this.insertions;
    // # Busy # SetType: Slanted, Boxed, Stepped, Periodic v0.38 posponed
    var underStepped = (painterStepSize * this.mainScaleS) < (painterLength * this.mainScaleX);
    var boxStepped = (painterStepSize * this.mainScaleS) == (painterLength * this.mainScaleX);
    var overStepped = (painterStepSize * this.mainScaleS) > (painterLength * this.mainScaleX);
    // # Busy # There is a difference between A: hasNeatEnds or B: underStepped v0.39 posponed
    this.hasNeatEnds = (this.beginData[0][0] > 0 || this.endData[0][0] > 0);

    // Initiate pi/2, counters, trash can and intermediate message string
    this.orthoAngle = Math.PI/2;
    this.countEntities = 0;
    this.countSplines = 0;
    this.countFitpoints = 0;
    this.trashCan = [];    // create Empty trash collector Array
    var msg = "";

    // With grouped castings: collect all additions to one single operation
    if (this.groupCasts) startTransaction();
    // Intentional +TAB: Transaction Block

        // Process all selected entities
        for (var i=0; i<idn; i++) {    // Cycle through all selected entities
            var id = ids[i];    // Pick Qlist entry
            var entity = document.queryEntity(id);    // Pick entity

            // Limit selection
            if (
                !isLineEntity(entity) &&
                !isArcEntity(entity) &&
                !isCircleEntity(entity) &&
                !isEllipseEntity(entity) &&
                !isPolylineEntity(entity) &&
                !isSplineEntity(entity)
            ) {
                    // Push smothered warning to command history
                    if (!this.beVerbose) EAction.handleUserWarning(entityTypeToString(entity.getType(), false) + " skipped!");
                    continue;    // Skip to next selected entity
            }

            var shape = entity.castToShape();

            // Set isTanCloObj flag for all geometrically and tangentially Closed Forms
            //  (entity based)
            this.isTanCloObj = false;
            if (isPolylineEntity(entity) || isSplineEntity(entity)) {
                if (entity.isGeometricallyClosed()) {
                    if (isSplineEntity(entity)) {
                        this.isTanCloObj = entity.isPeriodic();
                    }
                    else if (isPolylineEntity(entity)) {    // Near tangentially connected
                        this.isTanCloObj = RMath.fuzzyAngleCompare(shape.getDirection1() + Math.PI, shape.getDirection2());
                    }
                }
            }
            else if (isCircleEntity(entity) || isFullEllipseEntity(entity)) {
                this.isTanCloObj = true;
            }

            // Determine casting type
            // # Issue Solved # Its far more faster to set how to cast on the selected
            //           base entity then to test how to cast every point casted.
            if (isLineShape(shape) || isArcShape(shape) || isSplineShape(shape)) this.castingType = 1;
            if (isPolylineShape(shape)) this.castingType = 3;
            // # Issue Fixed # No mathematical Closed Form. 
            // Is by serie expansion. Is very time consuming
            // # Workaround # Ellipses by casting on a Polyline approximateWithArcs
            // # Issue # Relative acuracy, sized with the magnetude
            if (isEllipseEntity(entity)) {
                var shape = shape.approximateWithArcs(128);
                this.castingType = 3;
                if (this.isTanCloObj) shape.convertToClosed;
            }
            if (this.isTanCloObj) this.castingType++;    // Promote type 1 > 2 || 3 > 4
            if (isCircleShape(shape)) this.castingType = 5;

//debugger;
            // Get entity length after possible approximateWithArcs
            this.objlength = shape.getLength();

            // Understepped on Open Form:
            // Reserve 1 painter Length, n.knots+1 times
            var chainsN = this.insertions + 1;
            var reservedLength = 0.00;
            if (!this.isTanCloObj) {
                if (underStepped) {
                    var reservedLength = chainsN * painterLength;
                }
            }

            // Fit a number of actual painter StepSize lengths
            if (this.isTanCloObj) {    // Closed Forms: +No apexes, +n.knots, 1 more as possible
                var fitN = Math.ceil(
                            (this.objlength - (knotlength * this.insertions * this.mainScaleX))
                            /    // Divided by
                            (painterStepSize * this.mainScaleS));
            }
            else {    // Open Forms: +Apexes, +n.knots, 1 more as possible
                var fitN = Math.ceil(
                            (this.objlength - ((reservedLength + appendiceslength) * this.mainScaleX))
                            /    // Divided by
                            (painterStepSize * this.mainScaleS));
                // Not forced to scale: ceil() - 1 = floor()
                if (!this.forcedDyn) fitN--;
            }

            // Fit a larger multiple of chainsN to neatly divide in subchains
            fitN = Math.trunc(Math.ceil(fitN / chainsN)) * chainsN;

            // Test for too small entities
            if (fitN <= 3) {
                // Push smothered warning to command history
                if (!this.beVerbose) EAction.handleUserWarning(entityTypeToString(entity.getType(), false) + " Too short. Skipped!");
                continue;    // Skip to next selected entity
            }

            // Test for too many castings
            if (fitN > 9999) {
                // Push smothered warning to command history
                if (!this.beVerbose) EAction.handleUserWarning("Too many castings. Skipped!");
                continue;    // Skip to next selected entity
            }

//debugger;
            // Dynamic scaler for continuous Closed Forms and forced to scale Open Forms
            var dynScale = 1.00;
            if (this.isTanCloObj) {    // Continuous Closed Forms: +No apexes, +n.Knots
                dynScale =
                    this.objlength
                    /    // Divided by sum
                    ((knotlength * this.insertions * this.mainScaleX)
                    + (fitN * painterStepSize * this.mainScaleS));
            }
            else {    // Open Forms: +Apexes, +n.Konts
                if (this.forcedDyn) {    // Forced to scale Open Forms
                    dynScale =
                        this.objlength
                        /    // Divided by sum
                        (((reservedLength + appendiceslength) * this.mainScaleX) +
                        (fitN * painterStepSize * this.mainScaleS));
                }    // Else: Dynamic scaler stays f1.00
            }

            // Set actual scalers
            this.scalerY = this.mainScaleY;
            if (this.preserveLook) this.scalerY = this.mainScaleY * dynScale;
            this.scalerS = this.mainScaleS * dynScale;
            this.scalerX = this.mainScaleX * dynScale;
            // Flipping is inverse scaling in Y
            if (this.flipPainter) this.mainScaleY = -(this.mainScaleY);

            // Push smothered dynScale info to command history
            msg = entityTypeToString(entity.getType(), false);
            msg += (dynScale == 1) ? " not dynScaled" : " dynScaled : " + dynScale;
            if (!this.beVerbose) EAction.handleUserInfo(msg);

            // Only overtepped painters have voids
            if (overStepped) {
                var painterVoid = (painterStepSize * this.scalerS) - (painterLength * this.scalerX);
            }
            else {
                var painterVoid = 0.00;
            }

            // Local surplus absolute:
            // Any chain can locally only end with one such void if any
            var surPlusLabs = painterVoid;

            // Global surplus absolute:
            // Any void left globally
            var surPlusGabs =
                    this.objlength
                    - ((reservedLength + appendiceslength) * this.scalerX)
                    - (fitN * painterStepSize * this.scalerS);

            // # Todo # Treat originally slanted with neat ends different
            if (!this.isTanCloObj) {
                if (!this.forcedDyn) {
                // Pro / Cons ??? v0.39 posponed
                // Aligned and exposing a void globally is automatic
                // Overstepped with Apex not
                }
            }

            // On Open Forms except centered:
            // In any chain fits one single Null-Length painter extra
            if (!this.isTanCloObj && (this.factorsurPlus != 0.50)) {
                if (painterLength == 0.00) {
                    fitN+= chainsN;
                    surPlusLabs -= painterVoid
                }
            }

            // On Open Forms a chain may expose a void globally
            // Except centered: There is no place for 2 more full painters
            if (!this.isTanCloObj) {
                if (!this.doBeginApex) {    // To do and existing
                    // Aligned at end may expose a void globally
                    if (this.factorsurPlus == 1.00) {
                        if ((surPlusGabs + surPlusLabs) * 1 > painterLength * this.scalerX) {
                            // Add extra
                            fitN ++;
                        }
                    }
                }
                if (!this.doEndApex) {    // To do and existing
                    // Aligned at start may expose a void globally
                    if (this.factorsurPlus == 0.00) {
                        if ((surPlusGabs + surPlusLabs) * 1 > painterLength * this.scalerX) {
                            // Add extra
                            fitN ++;
                        }
                    }
                }
            }

            // Cast reserved too if any
            if (reservedLength > 0.00) fitN += chainsN;

            // Initiate global collector
            this.dotsSplineData = [];    // Empty global fitpoint splines collector
            this.nodesPolyData = [];    // Empty global polylines collector

            // Aligning & Shifting
            if (!this.isTanCloObj) {    // All Open Forms aligning
                this.distFromStart = surPlusGabs * this.factorsurPlus;
            }
            else {    // All Closed Forms shifting by % off stepsize
                this.distFromStart = painterStepSize * this.scalerS * this.closeShift;
            }

            // With non-grouped casting: collects additions to one single operation every entity
            if (!this.groupCasts) startTransaction();
            // Intentional +TAB Transaction Block

                // Cast & morphs painters over Entities
                // ------------------------------------
                // Optionally cast begin apex painter On Open Forms
                if (this.doBeginApex) {    // To do and existing
                    if (!this.isTanCloObj) {    // On Open Forms
                        this.countFitpoints += this.castPainter(shape, this.beginApexData);
                        this.countSplines += this.beginApexData[0][0] - this.beginApexData[0][1];
                        this.distFromStart += this.beginApexData[0][2] * this.scalerX;
                    }
                }

                // Repeat chain + insertion over whole the entity
                // Runs at least once, no insertion on last run
                for (var m=0; m<=this.insertions; m++) {
                    // Sub-chain shifting
                    this.distFromStart += surPlusLabs * this.factorsurPlus / chainsN;

                    // Add begin painter if any on Open Forms
                    if (this.beginData[0][0] > 0) {
                        if (!this.isTanCloObj) {    // On Open Forms
                            this.countFitpoints += this.castPainter(shape, this.beginData);
                            this.countSplines += this.beginData[0][0] - this.beginData[0][1];
                            // Length=0=fixed!
                        }
                    }

                    // Cast main painters
                    var startedAt = this.distFromStart
                    var nMax = fitN / chainsN;
                    for (var n=1; n<=nMax; n++) {    // Zero omitted = virtual(n+1)
                        this.countFitpoints += this.castPainter(shape, this.mainData);
                        this.countSplines += this.mainData[0][0] - this.mainData[0][1];
                        this.distFromStart = startedAt + (n * painterStepSize * this.scalerS);
                    // Loop main painters
                    }

                    // Add end painter if any on Open Forms
                    if (this.endData[0][0] > 0) {
                        if (!this.isTanCloObj) {    // On Open Forms
                            this.distFromStart += (painterLength * this.scalerX) - (painterStepSize * this.scalerS);
                            this.countFitpoints += this.castPainter(shape, this.endData);
                            this.countSplines += this.endData[0][0] - this.endData[0][1];
                            // Length=0=fixed!
                        }
                    }

                    // Sub-chain shifting
                    this.distFromStart += surPlusLabs * (1-this.factorsurPlus) / chainsN

                    // Optionally cast insertion
                    // Except for the last loop
                    if (m < this.insertions) {
                        if (this.doInserts) {    // To do and existing
                            this.countFitpoints += this.castPainter(shape, this.knotData);
                            this.countSplines += this.knotData[0][0] - this.knotData[0][1];
                            this.distFromStart += this.knotData[0][2] * this.scalerX;
                        }
                    }
                // Loop chains
                }

                // Optionally cast end apex painter on Open Forms
                if (this.doEndApex) {    // To do and existing
                    if (!this.isTanCloObj) {    // On Open Forms
                        this.distFromStart = this.objlength
                                             - (surPlusGabs * (1-this.factorsurPlus))
                                             - (this.endApexData[0][2] * this.scalerX);
                        this.countFitpoints += this.castPainter(shape, this.endApexData);
                        this.countSplines += this.endApexData[0][0] - this.endApexData[0][1];
                        //this.distFromStart += this.endApexData[0][2] * this.scalerX;
                    }
                }

                // Cast global splines if any, open/closed on open/Closed Forms
                var jn = this.dotsSplineData.length;
                for (var j=0; j<jn; j++) {    // Cycle through all global splines
                    addSpline(this.dotsSplineData[j], this.isTanCloObj);
                }
                this.countSplines += jn;

                // Cast global polys if any, open/closed on open/Closed Forms
                //var jn = this.nodesPolyData.length;

            // Intentional -TAB Transaction Block
            // Optionally parse all additions as one single operation every entity
            if (!this.groupCasts) {
                var castResult = endTransaction();    // No error if without any
                if (castResult.isFailed()) {    // Casting failed
                    break;    // Break out selected entities loop 
                }
                // # Issue # Not showing castings one by one
                if (i != idn - 1) {    // # workaround # Is slow: skip: only one and last
                    disableInput();
                    update();    // Fix to show progress sequentially
                    // # Issue # Flashing
                    enableInput();    // # Issue # Deprecated!! Meaning???
                }
            }

            if (this.delEntity) this.trashCan.push(id);
            this.countEntities++;
        // Loop all selected entities
        }

    // Intentional -TAB: Transaction Block
    // Optionally parse all additions as one single operation
    if (this.groupCasts) var castResult = endTransaction();    // No error if without any

    // Optionally empty trash can
    if (this.delEntity) {
        var jn = this.trashCan.length;
        if (jn > 0) {
            op = new RDeleteObjectsOperation();
            for (var j=0; j<jn; j++) {    // Cycle through all necessary deletions
                // query the stored entity copy back from the document
                // and use this to delete the entity
                op.deleteObject(document.queryEntity(this.trashCan[j]));
            }
            di.applyOperation(op);
    // # ToDo # On groupCasts transfer selection to casted v0.41?
    // # Issue # Selecting too many can greatly slow down Qcad
        }
    }

    // Push summary to command history
    if (this.countEntities == 0 || castResult.isFailed()) {
        // With none
        EAction.handleUserMessage("Nothing casted, command terminated");
    }
    else {    // With any
        if (!this.beVerbose) {    // Smothered
            EAction.handleUserInfo("- - - = * = - - -");
            msg = "Casted " + this.countSplines + " splines by " + this.countFitpoints + " fitpoints";
            EAction.handleUserInfo(msg);
            msg = "over " + this.countEntities;
            if (this.countEntities == idn) {
                msg += (this.countEntities > 1) ? " entities" : " entity";
            }
            else {
                msg += " of " + idn;
                msg += (idn > 1) ? " entities" : " entity";
            }
            msg += " as ";
            if (this.countEntities == 1 || this.groupCasts ) {
                msg += "one addition";
            }
            else {
                msg += this.countEntities;
                msg += (this.countEntities > 1) ? " addition" : " additions";
            }
            EAction.handleUserInfo(msg);
            if (this.delEntity && jn > 0) {
                msg = "Deleted " + jn;
                msg += (jn > 1) ? " entities" : " entity";
                msg += " as one deletion";
                EAction.handleUserInfo(msg);
            }
        }
        // Push command end message to command history
        msg = "Casting complete (" + (Date.now() - crono) + "ms)";
        EAction.handleUserMessage(msg);
    }

//    enableInput();
    this.terminate();
    return;
};

/**
 * Casts Painter local / global spline fitpoints.
 *
 * \return Number of casted fitpoints.
 */
// Sub-script of FlexPainter.beginEvent
FlexPainter.prototype.castPainter = function (shape, splineData) {
    // faster: local const  // faster: at top
    const objlength = this.objlength;
    const distFromStart = this.distFromStart;
    const skn = splineData[0][0];    // Number of entries in the dataset
    var countFitpoints = 0;

    // Cycle through all painter splines
    // # ToDo # Expand definition types: Point, Line, ...
    for (var sk=1; sk<=skn;sk++) {    // Zero omitted = header
        var newSplineData = [];    // Empty local fitpoint collector
        var partData = splineData[sk];   // Pick painter spline
        var nominator = partData[0];
        var dotStringN = (nominator == "spline") ? 0 : 1 * nominator.slice(4);    // Isolate dot string number
        // # ToDo # Expand nominator extraction
        var kn = partData.length - 1;    // Number of painter spline fitpoints
        countFitpoints += kn;

        // Cycle through all painter spline fitpoints
        for (var k=1; k<=kn;k++) {    // Zero omitted = header
            var offsetAtDist = partData[k][0];
            var scaledOffsetSize = partData[k][1] * this.scalerY;
            var activeDist = distFromStart + (offsetAtDist * this.scalerX);

            // Flaws with POLYLINES are getting pritty annoying
            //-------------------------------------------------
            // Catch castings nearly at entity length, copy at case4
            if (RMath.fuzzyCompare(activeDist, objlength, RS.PointTolerance)) {
                //activeDist = objlength;
                //qDebug("NearEnderCastpoint");
                //EAction.handleUserWarning("NearEnderCastpoint");
            };
            // Catch end crossers
            if (activeDist > objlength) {
                //activeDist = objlength;
                if (!this.isTanCloObj) {
                    qDebug("TooFarCaster");
                    if (!this.beVerbose) EAction.handleUserWarning("TooFarCaster");
                }
            };
            //-------------------------------------------------

            // Use different methods to determine castPoint & castAngle
            switch(this.castingType) {
                case 1:
                    // Lines, Arcs & open Splines simply by castPoint on shape
                    // Lines, Arcs No hard ending. Splines stiff ending, avoid crossing
                    // # Issue # IS NOT accurate on Splines, flaw larger rather near an endings?!
                    var castPoint = shape.getPointWithDistanceToStart(activeDist);
                    // # Issue Fixed # Geo closed spline: Cast on begin point has angle of ending
                    if (RMath.fuzzyCompare(activeDist, 0.00, RS.PointTolerance)) {
                        // # workaround # Offset a tiny bit from start
                        // # Issue # Fixing solution slows center loop back down
                        var castAngle = shape.getAngleAt(2*RS.PointTolerance, RS.FromStart);
                        qDebug("Type1BeginCastFix");
                        if (!this.beVerbose) EAction.handleUserWarning("Type1BeginCastFix");
                    }
                    else { 
                        var castAngle = shape.getAngleAt(activeDist, RS.FromStart);
                    }
                    break;
                case 2:
                    // Continuous closed Splines simply by castPoint on shape.
                    // isTanCloObj, simply restart from zero
                    var castPoint = shape.getPointWithDistanceToStart(activeDist % objlength);
                    var castAngle = shape.getAngleAt(activeDist, RS.FromStart);
                    break;
                case 3:
                    // Open Polyline & EllipseArc by castPoint on ClosestSegment
                    // Hard ending, avoid crossing
                    // # Issue Fixed # Geo closed poly: Cast on end point has angle of beginning (lines!arcs?)
                    //           As does casting on all vertices = angle at next segment
                    if (RMath.fuzzyCompare(activeDist, objlength, RS.PointTolerance)) {
                        // # workaround # Offset a tiny bit from end
                        // # Issue # This IS NOT a tiny bit. SEE also case 4
                        // # Issue # Fixing solution slows center loop back down
                        qDebug("Type3HugeEndCastFix");
                        if (!this.beVerbose) EAction.handleUserWarning("Type3HugeEndCastFix");
                        var castPoint = shape.getPointWithDistanceToStart(objlength - (2000*RS.PointTolerance));
                    }
                    else {
                        var castPoint = shape.getPointWithDistanceToStart(activeDist);
                    }
                    var segmentShape = shape.getSegmentAt(shape.getClosestSegment(castPoint));
                    var castAngle = segmentShape.getAngleAtPoint(castPoint);
                    break;
                case 4:
                    // Continuous closed Polyline by castPoint on ClosestSegment.
                    // isTanCloObj, simply restart from zero
                    // # Issue Fixed # Freak BUG casting non dynscaled very near endpoint eg 4e-14 @453.25--7166 on 453.25--717 long
                    // Fixed with Catching ClPolyNearEnder

                    if (RMath.fuzzyCompare(activeDist, objlength, RS.PointTolerance)) {
                        // # workaround # Fix at full lenght
                        // # Issue # Fixing solution slows center loop back down
                        activeDist = objlength;
                        qDebug("ClPolyNearEnderFix");
                        if (!this.beVerbose) EAction.handleUserWarning("ClPolyNearEnderFix");
                    };
                    var castPoint = shape.getPointWithDistanceToStart(activeDist % objlength);
                    var segmentShape = shape.getSegmentAt(shape.getClosestSegment(castPoint));
                    var castAngle = segmentShape.getAngleAtPoint(castPoint);
                    break;
                case 5:
                    // Circles by equation for castPoint & castAngle.
                    // Circles have no ending
                    var centerAngle = activeDist / shape.getRadius();
                    var castPoint = shape.getPointAtAngle(centerAngle);
                    var castAngle = centerAngle + this.orthoAngle;
                    break;
                default:
                    // Should not occur, good practices to include
            } 

            // Push casted fitpoint to collectors
            var castVector = new RVector(castPoint.x - Math.sin(castAngle) * scaledOffsetSize,
                                         castPoint.y + Math.cos(castAngle) * scaledOffsetSize);
            // Catch casting null-vectors
            if (isNull(castVector)) {
                qDebug("Casting Null-Vector");
                EAction.handleUserWarning("Casting Null-Vector");
                debugger;    // Faulty! Halt in debugger mode
            }
            // # ToDo # Expand definition types: Point, Line, ...
            if (nominator == "spline") {    // Push to local splines
                newSplineData.push(castVector);
            }
            else {    // Push to global splines
                // # Issue # Again, forgot why not push versus add past end ??? Array thing
                if (this.dotsSplineData.length == dotStringN) {    // Start a new dotString fitpoint collector
                    this.dotsSplineData[this.dotsSplineData.length] = [castVector];
                }
                else {    // Add to existing dotString fitpoint collector
                this.dotsSplineData[dotStringN][this.dotsSplineData[dotStringN].length] = castVector;
                }
            }
        // Loop all painter spline fitpoints
        }

        // Cast open spline with collected casted fitpoints
        // # Issue Solved # Is no error with empty newSplineData, will lead to AppCrash
        if (nominator == "spline") addSpline(newSplineData, false);
    // Loop all painter splines
    }
    return countFitpoints;
};


/**
 * Shows the painter dialog.
 *
 * \return True if the dialog was accepted, false if it was canceled.
 */
// Based on the scaling dialog Widget
FlexPainter.prototype.showDialog = function() {
    var dialog = WidgetFactory.createDialog(FlexPainter.includeBasePath, "FlexPainterDialog.ui");
    var widgets = getWidgets(dialog);

    // Load persistent dialog entries
    WidgetFactory.restoreState(dialog);

    // Detect first load: zero-width never makes sense
    if (widgets["MainWidth"].getValue() == 0.00) {
        // Painters, Knot, Icons and Prefs are as default
        // Initiate default painter set 'CurtainRope'
        this.PainterSetInit("CurtainRope");
        // Set an appropriate default painter width
        switch(this.units) {
            case 4:    // Millimeter and all not specified (Metric)
            default:
                widgets["MainWidth"].setValue(10.00);
                break;
            case 1:    // 1 Inch 2.54cm >> 10cm
                widgets["MainWidth"].setValue(3.94);
                break;
            case 2:    // 1 Foot 3.048dm >> 10dm
                widgets["MainWidth"].setValue(3.28);
                break;
            case 3:    // 1 Mile 1.609km >> 10km
                widgets["MainWidth"].setValue(6.22);
                break;
            case 8:    // 1 Microinch 25.4nm >> 100 nm
                widgets["MainWidth"].setValue(3.94);
                break;
            case 9:    // 1 Mil 25.4micron >> 100 micron
                widgets["MainWidth"].setValue(3.94);
                break;
            case 10:    // 1 Yard 0.9144m >> 10m
                widgets["MainWidth"].setValue(10.94);
                break;
        }
        widgets["UnitsLabel"].text = RUnit.unitToName(this.units);
        qDebug("First Use, " + RUnit.unitToName(this.units));
        // 'PreserveLook' is default set
        // Adapt Main Length & Main Step to painter set
        this.mainScaleY = widgets["MainWidth"].getValue() / 9;
        widgets["MainLengthX"].setValue(this.mainData[0][2] * this.mainScaleY);
        widgets["MainStepX"].setValue(this.mainData[0][3] * this.mainScaleY);
    }
    else {    // Not the first load
        if (widgets["UnitsLabel"].text != RUnit.unitToName(this.units)) {    // former used units differs
            qDebug("ReCalcPersistentDims");
            // # ToDo # Adapt dimensions appropriately.
            // !!RUnit: use correct Type or textual!!
            // Both metric >> do nothing
            // (RUnit.isMetric(widgets["UnitsLabel"].text) = RUnit.isMetric(this.units))
            // else >>
            // correction = convert (double value, RS::Unit source, RS::Unit dest)
            // widgets["MainWidth"].setValue = correction * widgets["MainWidth"].getValue
            // widgets["MainLengthX"].setValue = correction * widgets["MainWidth"].getValue
            // widgets["MainStepX"].setValue = correction * widgets["MainWidth"].getValue
        }



        // # Workaround # On other then first use initiate fixed painter set for now
            // PainterSetInit will run from initWidgets
//            this.PainterSetInit("Whatever, is forced");
            // Adapt Main Length & Main Step to painter set
            this.mainScaleY = widgets["MainWidth"].getValue() / 9;
            // # Busy # Surpress here so entries are persistent
            // Then entries can not reflect set change. Hit PreserveLook twice
//            widgets["MainLengthX"].setValue(this.mainData[0][2] * this.mainScaleY);
//            ["MainStepX"].setValue(this.mainData[0][3] * this.mainScaleY);



    }

    // Initiate the dialog change functions
    // # Issue Solved # Immediately active, but not working
    // # Workaround # Firts set values then activate the functions
    this.initWidgets(dialog);

    // # Issue # Dialog pops up with lots of decimals now
    // With the dialog change functions active less are displayed
    // RMathLineEdit is with precision 6 and displaying is java related

    do {    // Repeat untill entries are flawless
    // # Issue # Simple & Dirty! Surely there are neater ways to do this
        var errmsg = undefined;

        // Activate FlexPainter dialog modal
        if (!dialog.exec()) {
            dialog.destroy();
            EAction.activateMainWindow();
            return false;   // Dialog rejected
        }

        // Retrieve user scaling preferences
        // Get and validate mainScaleY entry
        this.mainScaleY = Math.abs(widgets["MainWidth"].getValue()) / 9;
        if (this.mainScaleY > 111) {    // Upper limit
            this.mainScaleY = 111;
            errmsg = "! Width crossed upper limit !";
        }
        if (this.mainScaleY < 0.01) {    // Lower limit
            this.mainScaleY = 0.01;
            errmsg = "! Width crossed lower limit !";
        }
        // Push width error warning to command history
        if (!isNull(errmsg)) {
            EAction.handleUserWarning(errmsg);
            // re-parse to UI
            widgets["MainWidth"].setValue(this.mainScaleY * 9);
            errmsg = "!";
        }

        // Get and validate MainPainterBox entries
        if (widgets["PreserveLook"].checked===true) {
            this.preserveLook = true;
            // Preserve the look with uniform scaling
            this.mainScaleX = this.mainScaleY;
            this.mainScaleS = this.mainScaleY;
        }
        else {    // Not checked
            this.preserveLook = false;
            // Get and validate mainScaleX entry
            // Painter set can have main painter length zero, avoid Div0
            if (this.mainData[0][2] == 0.00){    
                this.mainScaleX = 1.00;    // Just to avoid zero, Div0 later
                if (widgets["MainLengthX"].getValue() > 0.00) {
                    errmsg = "! Is zero-length painter !";
                }
            }
            else {
                this.mainScaleX = Math.abs(widgets["MainLengthX"].getValue()) / this.mainData[0][2];
            }
            if (this.mainScaleX > 111) {    // Upper limit
                this.mainScaleX = 111;
                errmsg = "! Length crossed upper limit !";
            }
            if (this.mainScaleX < 0.01) {    // Lower limit
                this.mainScaleX = 0.01;
                errmsg = "! Length crossed lower limit !";
            }
            // Push any length error warning to command history
            if (!isNull(errmsg)) {
                if (errmsg != "!") {
                    EAction.handleUserWarning(errmsg);
                    // re-parse to UI
                    widgets["MainLengthX"].setValue(this.mainData[0][2] * this.mainScaleX);
                errmsg ="!"
                }
            }

            // Get and validate mainScaleS entry
            this.mainScaleS = Math.abs(widgets["MainStepX"].getValue()) / this.mainData[0][3];
            if (this.mainScaleS > 111) {    // Upper limit
                this.mainScaleS = 111;
                errmsg = "! Stepsize crossed upper limit !";
            }
            if (this.mainScaleS < 0.01) {    // Lower limit
                this.mainScaleS = 0.01;
                errmsg = "! Stepsize crossed lower limit !";
            }
            // Push stepsize error warning to command history
            if (!isNull(errmsg)) {
                if (errmsg != "!") {
                    EAction.handleUserWarning(errmsg);
                    // re-parse to UI
                    widgets["MainStepX"].setValue(this.mainData[0][3] * this.mainScaleX);
                    errmsg ="!"
                }
            }
        }
    // Loop untill entries are flawless
    }
    while (!isNull(errmsg));

    // Retrieve user Open Forms shifting preference (0-100)
    this.closeShift = widgets["CloseShift"].value / 100;

    // Retrieve user casting preferences
    this.flipPainter = (widgets["FlipPainter"].checked===true);
    this.groupCasts = (widgets["GroupCasts"].checked===true);
    this.doBeginApex = (widgets["DoBeginApex"].checked===true);
        // Only do if exist (double lock checked/enabled)
        if (this.beginApexData[0][0] == 0) this.doBeginApex = false;
    this.doEndApex = (widgets["DoEndApex"].checked===true);
        // Only do if exist (double lock checked/enabled)
        if (this.endApexData[0][0] == 0) this.doEndApex = false;
    this.doInserts = (widgets["DoInserts"].checked===true);
        // Only do if exist (double lock checked/enabled)
        if (this.knotData[0][0] == 0) this.doInserts = false;
    // # ToDo # Temporary fixed for testing AND should be zero with false doInserts
    this.insertions = 0;
    this.beVerbose = (widgets["BeVerbose"].checked===true);
    this.delEntity = (widgets["DelEntity"].checked===true);

    // Retrieve user Open Forms scaling preferences
    this.factorsurPlus = 0.00;
    if (widgets["ForcedDyn"].checked===true) {
        this.forcedDyn = true;    // And no surPlusGabs length
    }
    else {    // Not forced to scale dynamically >> Some surPlusGabs length
        if (widgets["AlignCenter"].checked===true) {
            this.factorsurPlus = 0.50;    // Start with half the surPlusGabs length
        }
        else if (widgets["AlignEnd"].checked===true) {
            this.factorsurPlus = 1.00;    // Start with full surPlusGabs length
        }    // Else ("AlignStart") // Start with 0.00
    }

    // Retrieve user painter set selection
    // Both text based
    this.painterSet = widgets["PainterSelection"].currentText;
    this.painterKnot = widgets["KnotSelection"].currentText;

    // Push smothered painter summary to command history
    if (!this.beVerbose) {
        EAction.handleUserInfo("Painter set: " + this.painterSet);
        EAction.handleUserInfo("mainScale Widths: " + this.mainScaleY);
        EAction.handleUserInfo("mainScale Lengths: " + this.mainScaleX);
        EAction.handleUserInfo("mainScale Steps: " + this.mainScaleS);
        EAction.handleUserInfo("- - - = * = - - -");
    }

    // Save persistent dialog entries and finish
    widgets["BeVerbose"].checked = true;    // Super persistent checked
    widgets["DelEntity"].checked = false;    // Super persistent unchecked
    WidgetFactory.saveState(dialog);
    dialog.destroy();
    return true;   // Dialog accepted
};

/**
 * Initiate the FlexPainter dialog change functions
 */
// Based on the scaling dialog Widget
FlexPainter.prototype.initWidgets = function(widget, postfix) {

    if (isNull(postfix)) {    // Unknown purpose
        postfix = "";
    }


    // HERE IN SCOPE
    var widgets = getWidgets(widget);
    // Forgot to note where the analogue came from...
    var test1 = this.PainterSetInit(widgets["PainterSelection"].currentText);
    var test2 = this.mainData[0][2];    // base9 painter length
    var test3 = this.mainData[0][3];    // base9 painter stepsize
    var test4 = (this.beginApexData[0][0] > 0);    // Set has beginApexData
    var test5 = (this.endApexData[0][0] > 0);    // Set has endApexData
    var test6 = (this.knotData[0][0] > 0);    // Set has knotData
    // From this point on STATIC, most obviously
//debugger;
    // For on change functions that follow below:
    // # Issue # PainterInit function out of scope as attempts to callback
    // # Issue # Painter data out of scope
    // # Issue # test 1-5 out of scope
    // Tried (((this.)FlexPainter.)(this.))xyzxyz
    // (this. is presumebly just not more then the change function itself)
    // Testing complicated by modal nature of widget
    // FlexPainter.includeBasePath is in scope
    // Is not from inside FlexPainter.prototype.xyz function
    // Stuck, help needed
    // Found more:
    //    !!! Ucs,js has this.xyz spanning the script including slotchange != connect
    //    .blockSignals(true/false) when acting on widget objects
    //    HatchDialog.js is maybe a beter base???

    // On PainterSelection change:
    // Continuously: adapt Preview icons
    //               re-initiate painter set & Prefs
    //               parse MainWidth to MainPainterBox
    widgets["PainterSelection"].currentTextChanged.connect(
                function(currentText) {
                    // # ToDo # Check set data named 'currentText' and report if none
                    // # ToDo # Hardcoded "CurtainRope" vs currentText hereunder. Only one icon set yet
                    widgets["PreviewPainters"].icon = new QIcon(FlexPainter.includeBasePath + "/" + "CurtainRope" + ".svg");
                    widgets["PreviewKnot"].icon = new QIcon(FlexPainter.includeBasePath + "/" + "CurtainRope" + "Knot" +".svg");
                    // # ToDo # re-initiate painter set
                    // # Issue # FlexPainter.PainterSetInit() out of scope
//debugger;
                    // Adapt for Apexes and Insertions
                    // # Issue # this.mainData, test4-6 out of scope ... aren't set proper
                    widgets["DoBeginApex"].enabled = test4; // test4-6 are booleans
                    widgets["DoEndApex"].enabled = test5;
                    widgets["DoInserts"].enabled = test6;

                    // Width remains, force MainPainterBox to base9
                    widgets["PreserveLook"].checked = false
                    widgets["PreserveLook"].checked = true
                        // OR ( # Issue # Dirty! Toggle twice to get the correct state and actions)
                        // The same actions as in widgets["PreserveLook"].toggled.connect when ischecked
                });

    // On MainWidth change:
    // Continuously dismiss negative entries
    // With PreserveLook continuously parse entries in MainWidth to MainPainterBox
    widgets["MainWidth"].textChanged.connect(
                function(text) {
                    // Don't allow entry to start with 'minus'
                    widgets["MainWidth"].text = text.replace(/^-/, "");
                    // Parse MainWidth to MainPainterBox base9
                    if (widgets["PreserveLook"].checked===true) {
                        var scaleW = text / 9;
                        // # Issue # this.mainData, test2-3 out of scope ... or not?
                        widgets["MainLengthX"].text = test2 * scaleW;
                        widgets["MainStepX"].text = test3 * scaleW;
                    }
                });

    // On PreserveLook toggle:
    // # Issue # Difference between being toggled (=changeover) and to be toggled (=changeover to on)
    // Continuously: adapt PreserveLook icon
    //               when checked parse MainWidth to MainPainterBox base9
    widgets["PreserveLook"].toggled.connect(
                function(checked) {
                    if (checked) {    // Adapt icon
                        widgets["PreserveLook"].icon = new QIcon(FlexPainter.includeBasePath + "/PreserveLookOn" + postfix + ".svg");
                    }
                    else {
                        widgets["PreserveLook"].icon = new QIcon(FlexPainter.includeBasePath + "/PreserveLookOff" + postfix + ".svg");
                    }
                    if (checked) {    // Toggled to checked: Parse MainWidth to MainPainterBox
                        var scaleW = widgets["MainWidth"].text / 9;
                        // # Issue # this.mainData, test2-3 out of scope ... or not?
                        widgets["MainLengthX"].text = test2 * scaleW;
                        widgets["MainStepX"].text = test3 * scaleW;
                    }
                    // Adapt for Null-Length painter
                    if (test2 == 0.00) {
                        widgets["LabelLength"].enable = false;
                        widgets["MainLengthX"].enable = false;
                    }
                    else {
                        widgets["LabelLength"].enable = true;
                        widgets["MainLengthX"].enable = true;
                    }
                });

    // On MainLength change:
    // Continuously dismiss negative entries
    widgets["MainLengthX"].textChanged.connect(
                function(text) {
                    // Don't allow entry to start with 'minus'
                    widgets["MainLengthX"].text = text.replace(/^-/, "");
                });

    // On MainStepX change:
    // Continuously dismiss negative entries
    widgets["MainStepX"].textChanged.connect(
                function(text) {
                    // Don't allow entry to start with 'minus'
                    widgets["MainStepX"].text = text.replace(/^-/, "");
                });

};

/**
 * Initiate Painter set data.
 *
 * \return painterSet done for now, never fails to get a set # ToDo #
 */
// Sub-script of FlexPainter.beginEvent
FlexPainter.prototype.PainterSetInit = function(painterSet) {

    // # Busy # 12 Hardcoded painter set for testing purpose
            // Closed Forms align too...
        painterSet = "CurtainRope"; // Passed. Open Forms step > length: Apex detached
        //painterSet = "OrthogonalBar"; // Passed
        //painterSet = "OrthogonalPointsGlobal"; // Passed
        //painterSet = "SinusoidWave"; // Passed
        //painterSet = "SinusoidWaveGlobal"; // Similar
        //painterSet = "3PhaseWaveGlobal"; // Similar
        //painterSet = "TriangleWave"; // Similar
        //painterSet = "SawtoothWave"; // Similar
        //painterSet = "SqaureWave"; // Similar
        //painterSet = "SteppedWave"; // Similar
        //painterSet = "SqauredShape"; // Passed
        //painterSet = "SqauredShapeFlexing"; // Similar

        EAction.handleUserMessage("Forced: " + painterSet);    //Testing
        qDebug("Forced: " + painterSet);    //Testing

// ** HARDCODED in Arrays, works similar as splineEntity.objects **
// ** ADJUSTABLE SPLINE DATA SETS **    Painters splines fitpoint vectors
    switch(painterSet) {    // * = fixed for all
        case "CurtainRope": // Change vertical to Line
        default:
        this.beginApexData = [[29,0,12.6000]];    // *beginApexData.count, *Dots.count, *painterLength, ....... 
            this.beginApexData.push(["spline",[11.3175,4.1786],[11.2744,4.3393],[11.1568,4.4569],[10.9961,4.5000],[10.8354,4.4569],[10.7177,4.3393],[10.6746,4.1786]]);
            this.beginApexData.push(["spline",[8.1032,4.1786],[7.5600,-2.8607]]);
            this.beginApexData.push(["spline",[10.0318,4.1786],[9.9887,4.3393],[9.8711,4.4569],[9.7104,4.5000],[9.5496,4.4569],[9.4320,4.3393],[9.3889,4.1786]]);
            this.beginApexData.push(["spline",[8.7461,4.1786],[8.7030,4.3393],[8.5854,4.4569],[8.4246,4.5000],[8.2639,4.4569],[8.1463,4.3393],[8.1032,4.1786]]);
            this.beginApexData.push(["spline",[8.7461,4.1786],[8.1032,-4.1786]]);
            this.beginApexData.push(["spline",[11.9604,4.1786],[11.9604,-2.9663]]);
            this.beginApexData.push(["spline",[11.3175,-4.1786],[11.3606,-4.3393],[11.4782,-4.4569],[11.6389,-4.5000],[11.7996,-4.4569],[11.9173,-4.3393],[12.6643,-3.5281]]);
            this.beginApexData.push(["spline",[12.6643,-3.5281],[11.9604,-2.9663]]);
            this.beginApexData.push(["spline",[1.0286,-5.3036],[2.3143,-4.4679],[4.2429,-3.6321],[5.8500,-3.3750]]);
            this.beginApexData.push(["spline",[6.8143,3.8571],[6.4929,3.8571],[2.9571,5.1429],[0.7071,6.4286],[0.0855,3.4413],[0.0032,0.0000],[0.0855,-3.4413],[0.7071,-6.4286],[2.9571,-5.1429],[6.4929,-3.8571],[6.8143,-3.8571]]);
            this.beginApexData.push(["spline",[0.9643,-4.2107],[1.9929,-3.6321],[3.2786,-3.1821],[4.5643,-2.9571]]);
            this.beginApexData.push(["spline",[1.0286,-3.1500],[2.0571,-2.7321],[2.8286,-2.5393]]);
            this.beginApexData.push(["spline",[6.8143,-3.8571],[6.8143,3.8571]]);
            this.beginApexData.push(["spline",[7.4604,-4.1786],[7.4604,4.1786]]);
            this.beginApexData.push(["spline",[11.3175,4.1786],[10.6746,-4.1786]]);
            this.beginApexData.push(["spline",[10.0318,4.1786],[9.3889,-4.1786]]);
            this.beginApexData.push(["spline",[11.8607,2.8607],[11.3175,-4.1786]]);
            this.beginApexData.push(["spline",[8.1032,-4.1786],[8.0602,-4.3393],[7.9425,-4.4569],[7.7818,-4.5000],[7.6211,-4.4569],[7.5034,-4.3393],[7.4604,-4.1786]]);
            this.beginApexData.push(["spline",[10.6746,4.1786],[10.0318,-4.1786]]);
            this.beginApexData.push(["spline",[10.6746,-4.1786],[10.6316,-4.3393],[10.5139,-4.4569],[10.3532,-4.5000],[10.1925,-4.4569],[10.0748,-4.3393],[10.0318,-4.1786]]);
            this.beginApexData.push(["spline",[9.3889,-4.1786],[9.3459,-4.3393],[9.2282,-4.4569],[9.0675,-4.5000],[8.9068,-4.4569],[8.7891,-4.3393],[8.7461,-4.1786]]);
            this.beginApexData.push(["spline",[9.3889,4.1786],[8.7461,-4.1786]]);
            this.beginApexData.push(["spline",[11.9604,4.1786],[11.9173,4.3393],[11.7996,4.4569],[11.6389,4.5000],[11.4782,4.4569],[11.3606,4.3393],[11.3175,4.1786]]);
            this.beginApexData.push(["spline",[10.6746,4.1786],[10.6316,4.3393],[10.5139,4.4569],[10.3532,4.5000],[10.1925,4.4569],[10.0748,4.3393],[10.0318,4.1786]]);
            this.beginApexData.push(["spline",[9.3889,4.1786],[9.3459,4.3393],[9.2282,4.4569],[9.0675,4.5000],[8.9068,4.4569],[8.7891,4.3393],[8.7461,4.1786]]);
            this.beginApexData.push(["spline",[8.1032,4.1786],[8.0602,4.3393],[7.9425,4.4569],[7.7818,4.5000],[7.6211,4.4569],[7.5034,4.3393],[7.4604,4.1786]]);
            this.beginApexData.push(["spline",[8.7461,-4.1786],[8.7030,-4.3393],[8.5854,-4.4569],[8.4246,-4.5000],[8.2639,-4.4569],[8.1463,-4.3393],[8.1032,-4.1786]]);
            this.beginApexData.push(["spline",[11.3175,-4.1786],[11.2744,-4.3393],[11.1568,-4.4569],[10.9961,-4.5000],[10.8354,-4.4569],[10.7177,-4.3393],[10.6746,-4.1786]]);
            this.beginApexData.push(["spline",[10.0318,-4.1786],[9.9887,-4.3393],[9.8711,-4.4569],[9.7104,-4.5000],[9.5496,-4.4569],[9.4320,-4.3393],[9.3889,-4.1786]]);
        this.beginData = [[5,1,0.0000]];    // *beginData.count, *Dots.count, *painterLength, .......
            this.beginData.push(["spline",[0.0032,2.1445],[0.0032,-2.1502]]);
            this.beginData.push(["spline",[0.0032,-2.1502],[1.4047,-3.3837],[1.8885,-3.8107],[3.3783,-4.5514],[5.0389,-4.4480],[6.4254,-3.5281]]);
            this.beginData.push(["spline",[6.4254,-3.5281],[4.8198,-2.1109],[3.2143,-0.6936],[1.6088,0.7236],[0.0032,2.1445]]);
            this.beginData.push(["spline",[2.7321,-2.5714],[3.8912,-3.5028],[5.0389,-4.4480]]);
            this.beginData.push(["Dots00",[0.3214,4.4786]]);
        this.mainData = [[4,1,12.8571,6.4286,"Slanted"]];    // *mainData.count, *Dots.count, *painterLength, painterStepSize, painterType, .......
            this.mainData.push(["spline",[0.0032,3.5281],[2.0817,1.6934],[4.1601,-0.1413],[6.2386,-1.9760],[7.8333,-3.3837],[8.3170,-3.8107],[9.8069,-4.5514],[11.4675,-4.4480],[12.8539,-3.5281]]);
            this.mainData.push(["spline",[12.8539,-3.5281],[10.7787,-1.6898],[8.6970,0.1413],[6.8286,1.7957],[5.0221,3.3853],[4.5401,3.8107],[3.0503,4.5514],[1.3896,4.4480],[0.0032,3.5281]]);
            this.mainData.push(["spline",[9.1607,-2.5714],[10.3198,-3.5028],[11.4675,-4.4480]]);
            this.mainData.push(["Dots00",[1.4464,4.9607],[3.375,4.9607],[5.7857,4.1571],[6.75,4.4786]]);
        this.endData = [[12,1,0.0000]];    // *endData.count, *Dots.count, *painterLength, .......
            this.endData.push(["spline",[-6.4286,3.5281],[-4.8222,2.1102],[-3.2159,0.6922],[-1.6096,-0.7258],[-0.0032,-2.1437]]);
            this.endData.push(["spline",[-0.0032,-2.1437],[-0.0032,2.1437]]);
            this.endData.push(["spline",[-0.0032,2.1437],[-1.4097,3.3853],[-1.8917,3.8107],[-3.3815,4.5514],[-5.0422,4.4480],[-6.4286,3.5281]]);
            // - - - - CVH mark - - - - 
            this.endData.push(["spline",[-0.1284,0.0475],[-0.1284,0.0550],[-0.1246,0.0775],[-0.1146,0.0875],[-0.0896,0.0875],[-0.0796,0.0775],[-0.0759,0.0550],[-0.0759,0.0475]]);
            this.endData.push(["spline",[-0.0759,0.0475],[-0.1271,0.0300]]);
            this.endData.push(["spline",[-0.1271,0.0300],[-0.0771,0.0050]]);
            this.endData.push(["spline",[-0.0771,0.0050],[-0.0771,-0.0700]]);
            this.endData.push(["spline",[-0.0771,-0.0700],[-0.0896,-0.0700]]);
            this.endData.push(["spline",[-0.0896,-0.0200],[-0.1271,-0.0200]]);
            this.endData.push(["spline",[-0.1084,-0.0200],[-0.1084,-0.0450]]);
            this.endData.push(["spline",[-0.0896,-0.0450],[-0.1271,-0.0450]]);
            // - - - - CVH mark - - - - 
            this.endData.push(["Dots00",[-4.9821,4.9607],[-3.0536,4.9607],[-0.6429,4.1571]]);
        this.endApexData = [[25,1,12.8571]];    // *endApexData.count, *Dots.count, *painterLength, .......
            this.endApexData.push(["spline",[1.2825,-4.1786],[1.2825,4.1786]]);
            this.endApexData.push(["spline",[0.6396,-4.1786],[0.6396,2.9663]]);
            this.endApexData.push(["spline",[1.2825,4.1786],[1.2394,4.3393],[1.1218,4.4569],[0.9611,4.5000],[0.8004,4.4569],[0.6827,4.3393],[-0.0643,3.5281]]);
            this.endApexData.push(["spline",[-0.0643,3.5281],[0.6396,2.9663]]);
            this.endApexData.push(["spline",[5.7857,-3.8571],[6.1071,-3.8571],[9.6429,-5.1429],[11.8929,-6.4286]]);
            this.endApexData.push(["spline",[5.7857,3.8571],[6.1071,3.8571],[9.6429,5.1429],[11.8929,6.4286]]);
            this.endApexData.push(["spline",[12.5145,-3.4413],[11.2179,-2.9250],[9.8036,-2.5714]]);
            this.endApexData.push(["spline",[11.8929,6.4286],[12.8539,0.0000],[11.8929,-6.4286]]);
            this.endApexData.push(["spline",[12.3223,-4.4512],[10.9286,-3.7929],[9.6429,-3.2786],[8.0357,-2.9571]]);
            this.endApexData.push(["spline",[12.1096,-5.4547],[10.2857,-4.5000],[8.3571,-3.6643],[7.0714,-3.4071]]);
            this.endApexData.push(["spline",[5.1396,4.1786],[5.1396,-4.1786]]);
            this.endApexData.push(["spline",[3.8539,4.1786],[3.8539,-4.1786]]);
            this.endApexData.push(["spline",[4.4968,-4.1786],[4.5398,-4.3393],[4.6575,-4.4569],[4.8182,-4.5000],[4.9789,-4.4569],[5.0966,-4.3393],[5.1396,-4.1786]]);
            this.endApexData.push(["spline",[3.2111,-4.1786],[3.2541,-4.3393],[3.3718,-4.4569],[3.5325,-4.5000],[3.6932,-4.4569],[3.8109,-4.3393],[3.8539,-4.1786]]);
            this.endApexData.push(["spline",[2.5682,4.1786],[2.5682,-4.1786]]);
            this.endApexData.push(["spline",[1.9254,-4.1786],[1.9684,-4.3393],[2.0861,-4.4569],[2.2468,-4.5000],[2.4075,-4.4569],[2.5252,-4.3393],[2.5682,-4.1786]]);
            this.endApexData.push(["spline",[0.6396,-4.1786],[0.6827,-4.3393],[0.8004,-4.4569],[0.9611,-4.5000],[1.1218,-4.4569],[1.2394,-4.3393],[1.2825,-4.1786]]);
            this.endApexData.push(["spline",[4.4968,4.1786],[4.4968,-4.1786]]);
            this.endApexData.push(["spline",[1.9254,4.1786],[1.9684,4.3393],[2.0861,4.4569],[2.2468,4.5000],[2.4075,4.4569],[2.5252,4.3393],[2.5682,4.1786]]);
            this.endApexData.push(["spline",[3.2111,4.1786],[3.2541,4.3393],[3.3718,4.4569],[3.5325,4.5000],[3.6932,4.4569],[3.8109,4.3393],[3.8539,4.1786]]);
            this.endApexData.push(["spline",[4.4968,4.1786],[4.5398,4.3393],[4.6575,4.4569],[4.8182,4.5000],[4.9789,4.4569],[5.0966,4.3393],[5.1396,4.1786]]);
            this.endApexData.push(["spline",[3.2111,4.1786],[3.2111,-4.1786]]);
            this.endApexData.push(["spline",[1.9254,4.1786],[1.9254,-4.1786]]);
            this.endApexData.push(["spline",[5.7857,-3.8571],[5.7857,3.8571]]);
            this.endApexData.push(["Dots01",[-0.2327,-4.4996],[-1.5183,-5.6245],[-2.6432,-5.7852],[-4.5716,-5.4638],[-4.5716,-5.1424],[-2.6432,-5.3031],[-0.2327,-4.1782]]);
        this.knotData = [[0,0,0.0000,1,50.00]];    // *knotData.count, *Dots.count, *painterLength, insertions, insertPoints.......
        // # ToDo # 1insertions, insertPoint 50% with length zero to work without knots
        //    this.knotData.push([[0.0,0.0],[0.0,0.0],....]);
        break;
    case "OrthogonalBar": // Change vertical to Line
        this.beginApexData = [[0,0,0.0000]];    // *beginApexData.count, *Dots.count, *painterLength, .......
        this.beginData = [[0,0,0.0000]];    // *beginData.count, *Dots.count, *painterLength, .......
        this.mainData = [[1,0,0.0000,9.0000,"Stepped"]];    // *mainData.count, *Dots.count, *painterLength, painterStepSize, .......
            this.mainData.push(["spline",[0.0000,4.5000],[0.0000,-4.5000]]);
        this.endData = [[0,0,0.0000]];    // *endData.count, *Dots.count, *painterLength, .......
        this.endApexData = [[0,0,0.0000]];    // *endApexData.count, *Dots.count, *painterLength, .......
        this.knotData = [[0,0,0.0000,0,0.00]];    // *knotData.count, *Dots.count, *painterLength, insertions, insertPoints.......
        break;
    case "OrthogonalPointsGlobal":
        this.beginApexData = [[0,0,0.0000]];    // *beginApexData.count, *Dots.count, *painterLength, .......
        this.beginData = [[0,0,0.0000]];    // *beginData.count, *Dots.count, *painterLength, .......
        this.mainData = [[2,2,0.0000,9.0000,"Stepped"]];    // *mainData.count, *Dots.count, *painterLength, painterStepSize, .......
            this.mainData.push(["Dots00",[0.0000,4.5000]]);
            this.mainData.push(["Dots01",[0.0000,-4.5000]]);
        this.endData = [[0,0,0.0000]];    // *endData.count, *Dots.count, *painterLength, .......
        this.endApexData = [[0,0,0.0000]];    // *endApexData.count, *Dots.count, *painterLength, .......
        this.knotData = [[0,0,0.0000,0,0.00]];    // *knotData.count, *Dots.count, *painterLength, insertions, insertPoints.......
        break;
    case "SinusoidWave":
        this.beginApexData = [[0,0,0.0000]];    // *beginApexData.count, *Dots.count, *painterLength, .......
        this.beginData = [[0,0,0.0000]];    // *beginData.count, *Dots.count, *painterLength, .......
        this.mainData = [[2,0,9.0000,9.0000,"Boxed"]];    // *mainData.count, *Dots.count, *painterLength, painterStepSize, .......
            this.mainData.push(["spline",[0.0000,0.0000],[0.5000,1.5391],[1.0000,2.8925],[1.5000,3.8971],[1.7500,4.2286],[2.0000,4.4316],[2.2500,4.5000],[2.5000,4.4316],[2.7500,4.2286],[3.0000,3.8971],[3.5000,2.8925],[4.0000,1.5391],[4.5000,0.0000]]);
            this.mainData.push(["spline",[4.5000,0.0000],[5.0000,-1.5391],[5.5000,-2.8925],[6.0000,-3.8971],[6.2500,-4.2286],[6.5000,-4.4316],[6.7500,-4.5000],[7.0000,-4.4316],[7.2500,-4.2286],[7.5000,-3.8971],[8.0000,-2.8925],[8.5000,-1.5391],[9.0000,0.0000]]);
        this.endData = [[0,0,0.0000]];    // *endData.count, *Dots.count, *painterLength, .......
        this.endApexData = [[0,0,0.0000]];    // *endApexData.count, *Dots.count, *painterLength, .......
        this.knotData = [[0,0,0.0000,0,0.00]];    // *knotData.count, *Dots.count, *painterLength, insertions, insertPoints.......
        break;
    case "SinusoidWaveGlobal":
        this.beginApexData = [[0,0,0.0000]];    // *beginApexData.count, *Dots.count, *painterLength, .......
        this.beginData = [[0,0,0.0000]];    // *beginData.count, *Dots.count, *painterLength, .......
        this.mainData = [[2,2,9.0000,9.0000,"Boxed"]];    // *mainData.count, *Dots.count, *painterLength, painterStepSize, .......
            this.mainData.push(["Dots00",[0.0000,0.0000],[0.5000,1.5391],[1.0000,2.8925],[1.5000,3.8971],[1.7500,4.2286],[2.0000,4.4316],[2.2500,4.5000],[2.5000,4.4316],[2.7500,4.2286],[3.0000,3.8971],[3.5000,2.8925],[4.0000,1.5391],[4.5000,0.0000]]);
            this.mainData.push(["Dots00",[4.5000,0.0000],[5.0000,-1.5391],[5.5000,-2.8925],[6.0000,-3.8971],[6.2500,-4.2286],[6.5000,-4.4316],[6.7500,-4.5000],[7.0000,-4.4316],[7.2500,-4.2286],[7.5000,-3.8971],[8.0000,-2.8925],[8.5000,-1.5391],[9.0000,0.0000]]);
        this.endData = [[0,0,0.0000]];    // *endData.count, *Dots.count, *painterLength, .......
        this.endApexData = [[0,0,0.0000]];    // *endApexData.count, *Dots.count, *painterLength, .......
        this.knotData = [[0,0,0.0000,0,0.00]];    // *knotData.count, *Dots.count, *painterLength, insertions, insertPoints.......
        break;
    case "3PhaseWaveGlobal":
        this.beginApexData = [[0,0,0.0000]];    // *beginApexData.count, *Dots.count, *painterLength, .......
        this.beginData = [[0,0,0.0000]];    // *beginData.count, *Dots.count, *painterLength, .......
        this.mainData = [[8,8,9.0000,9.0000,"Boxed"]];    // *mainData.count, *Dots.count, *painterLength, painterStepSize, .......
            this.mainData.push(["Dots00",[0.0000,0.0000],[0.5000,1.5391],[1.0000,2.8925],[1.5000,3.8971],[1.7500,4.2286],[2.0000,4.4316],[2.2500,4.5000],[2.5000,4.4316],[2.7500,4.2286],[3.0000,3.8971],[3.5000,2.8925],[4.0000,1.5391],[4.5000,0.0000]]);
            this.mainData.push(["Dots00",[4.5000,0.0000],[5.0000,-1.5391],[5.5000,-2.8925],[6.0000,-3.8971],[6.2500,-4.2286],[6.5000,-4.4316],[6.7500,-4.5000],[7.0000,-4.4316],[7.2500,-4.2286],[7.5000,-3.8971],[8.0000,-2.8925],[8.5000,-1.5391],[9.0000,0.0000]]);
            this.mainData.push(["Dots01",[0.0000,3.8971],[0.5000,2.8925],[1.0000,1.5391],[1.5000,0.0000]]);
            this.mainData.push(["Dots01",[1.5000,0.0000],[2.0000,-1.5391],[2.5000,-2.8925],[3.0000,-3.8971],[3.2500,-4.2286],[3.5000,-4.4316],[3.7500,-4.5000],[4.0000,-4.4316],[4.2500,-4.2286],[4.5000,-3.8971],[5.0000,-2.8925],[5.5000,-1.5391],[6.0000,0.0000]]);
            this.mainData.push(["Dots01",[6.0000,0.0000],[6.5000,1.5391],[7.0000,2.8925],[7.5000,3.8971],[7.7500,4.2286],[8.0000,4.4316],[8.2500,4.5000],[8.5000,4.4316],[8.7500,4.2286],[9.0000,3.8971]]);
            this.mainData.push(["Dots02",[0.0000,-3.8971],[0.2500,-4.2286],[0.5000,-4.4316],[0.7500,-4.5000],[1.0000,-4.4316],[1.2500,-4.2286],[1.5000,-3.8971],[2.0000,-2.8925],[2.5000,-1.5391],[3.0000,0.0000]]);
            this.mainData.push(["Dots02",[3.0000,0.0000],[3.5000,1.5391],[4.0000,2.8925],[4.5000,3.8971],[4.7500,4.2286],[5.0000,4.4316],[5.2500,4.5000],[5.5000,4.4316],[5.7500,4.2286],[6.0000,3.8971],[6.5000,2.8925],[7.0000,1.5391],[7.5000,0.0000]]);
            this.mainData.push(["Dots02",[7.5000,0.0000],[8.0000,-1.5391],[8.5000,-2.8925],[9.0000,-3.8971]]);
            this.endData = [[0,0,0.0000]];    // *endData.count, *Dots.count, *painterLength, .......
        this.endApexData = [[0,0,0.0000]];    // *endApexData.count, *Dots.count, *painterLength, .......
        this.knotData = [[0,0,0.0000,0,0.00]];    // *knotData.count, *Dots.count, *painterLength, insertions, insertPoints.......
        break;
    case "TriangleWave":
        this.beginApexData = [[0,0,0.0000]];    // *beginApexData.count, *Dots.count, *painterLength, .......
        this.beginData = [[0,0,0.0000]];    // *beginData.count, *Dots.count, *painterLength, .......
        this.mainData = [[4,0,9.0000,9.0000,"Boxed"]];    // *mainData.count, *Dots.count, *painterLength, painterStepSize, .......
            this.mainData.push(["spline",[0.0000,0.0000],[2.2500,4.5000]]);
            this.mainData.push(["spline",[2.2500,4.5000],[4.5000,0.0000]]);
            this.mainData.push(["spline",[4.5000,0.0000],[6.7500,-4.5000]]);
            this.mainData.push(["spline",[6.7500,-4.5000],[9.0000,0.0000]]);
        this.endData = [[0,0,0.0000]];    // *endData.count, *Dots.count, *painterLength, .......
        this.endApexData = [[0,0,0.0000]];    // *endApexData.count, *Dots.count, *painterLength, .......
        this.knotData = [[0,0,0.0000,1,50.00]];    // *knotData.count, *Dots.count, *painterLength, insertions, insertPoints.......
        // # ToDo # 1insertions, insertPoint 50% with length zero to work without knots
        break;
    case "SawtoothWave": // Change vertical to Line
        this.beginApexData = [[0,0,0.0000]];    // *beginApexData.count, *Dots.count, *painterLength, .......
        this.beginData = [[0,0,0.0000]];    // *beginData.count, *Dots.count, *painterLength, .......
        this.mainData = [[4,0,9.0000,9.0000,"Boxed"]];    // *mainData.count, *Dots.count, *painterLength, painterStepSize, .......
            this.mainData.push(["spline",[0.0000,0.0000],[4.5000,4.5000]]);
            this.mainData.push(["spline",[4.5000,4.5000],[4.5000,0.0000]]);
            this.mainData.push(["spline",[4.5000,0.0000],[4.5000,-4.5000]]);
            this.mainData.push(["spline",[4.5000,-4.5000],[9.0000,0.0000]]);
        this.endData = [[0,0,0.0000]];    // *endData.count, *Dots.count, *painterLength, .......
        this.endApexData = [[0,0,0.0000]];    // *endApexData.count, *Dots.count, *painterLength, .......
        this.knotData = [[0,0,0.0000,0,0.00]];    // *knotData.count, *Dots.count, *painterLength, insertions, insertPoints.......
        break;
    case "SqaureWave": // Change to Line
        this.beginApexData = [[0,0,0.0000]];    // *beginApexData.count, *Dots.count, *painterLength, .......
        this.beginData = [[0,0,0.0000]];    // *beginData.count, *Dots.count, *painterLength, .......
        this.mainData = [[6,0,9.0000,9.0000,"Boxed"]];    // *mainData.count, *Dots.count, *painterLength, painterStepSize, .......
            this.mainData.push(["spline",[0.0000,0.0000],[0.0000,4.5000]]);
            this.mainData.push(["spline",[0.0000,4.5000],[4.5000,4.5000]]);
            this.mainData.push(["spline",[4.5000,4.5000],[4.5000,0.0000]]);
            this.mainData.push(["spline",[4.5000,0.0000],[4.5000,-4.5000]]);
            this.mainData.push(["spline",[4.5000,-4.5000],[9.0000,-4.5000]]);
            this.mainData.push(["spline",[9.0000,-4.5000],[9.0000,0.0000]]);
        this.endData = [[0,0,0.0000]];    // *endData.count, *Dots.count, *painterLength, .......
        this.endApexData = [[0,0,0.0000]];    // *endApexData.count, *Dots.count, *painterLength, .......
        this.knotData = [[0,0,0.0000,0,0.00]];    // *knotData.count, *Dots.count, *painterLength, insertions, insertPoints.......
        break;
    case "SteppedWave": // Change to Line
        this.beginApexData = [[0,0,0.0000]];    // *beginApexData.count, *Dots.count, *painterLength, .......
        this.beginData = [[0,0,0.0000]];    // *beginData.count, *Dots.count, *painterLength, .......
        this.mainData = [[22,0,9.0000,9.0000,"Boxed"]];    // *mainData.count, *Dots.count, *painterLength, painterStepSize, .......
            this.mainData.push(["spline",[0.0000,0.0000],[0.0000,1.5000]]);
            this.mainData.push(["spline",[0.0000,1.5000],[0.9000,1.5000]]);
            this.mainData.push(["spline",[0.9000,1.5000],[0.9000,3.0000]]);
            this.mainData.push(["spline",[0.9000,3.0000],[1.8000,3.0000]]);
            this.mainData.push(["spline",[1.8000,3.0000],[1.8000,4.5000]]);
            this.mainData.push(["spline",[1.8000,4.5000],[2.7000,4.5000]]);
            this.mainData.push(["spline",[2.7000,4.5000],[2.7000,3.0000]]);
            this.mainData.push(["spline",[2.7000,3.0000],[3.6000,3.0000]]);
            this.mainData.push(["spline",[3.6000,3.0000],[3.6000,1.5000]]);
            this.mainData.push(["spline",[3.6000,1.5000],[4.5000,1.5000]]);
            this.mainData.push(["spline",[4.5000,1.5000],[4.5000,0.0000]]);
            this.mainData.push(["spline",[4.5000,0.0000],[4.5000,-1.5000]]);
            this.mainData.push(["spline",[4.5000,-1.5000],[5.4000,-1.5000]]);
            this.mainData.push(["spline",[5.4000,-1.5000],[5.4000,-3.0000]]);
            this.mainData.push(["spline",[5.4000,-3.0000],[6.3000,-3.0000]]);
            this.mainData.push(["spline",[6.3000,-3.0000],[6.3000,-4.5000]]);
            this.mainData.push(["spline",[6.3000,-4.5000],[7.2000,-4.5000]]);
            this.mainData.push(["spline",[7.2000,-4.5000],[7.2000,-3.0000]]);
            this.mainData.push(["spline",[7.2000,-3.0000],[8.1000,-3.0000]]);
            this.mainData.push(["spline",[8.1000,-3.0000],[8.1000,-1.5000]]);
            this.mainData.push(["spline",[8.1000,-1.5000],[9.0000,-1.5000]]);
            this.mainData.push(["spline",[9.0000,-1.5000],[9.0000,0.0000]]);
        this.endData = [[0,0,0.0000]];    // *endData.count, *Dots.count, *painterLength, .......
        this.endApexData = [[0,0,0.0000]];    // *endApexData.count, *Dots.count, *painterLength, .......
        this.knotData = [[0,0,0.0000,0,0.00]];    // *knotData.count, *Dots.count, *painterLength, insertions, insertPoints.......
        break;
    case "SqauredShape":
        this.beginApexData = [[0,0,0.0000]];    // *beginApexData.count, *Dots.count, *painterLength, .......
        this.beginData = [[0,0,0.0000]];    // *beginData.count, *Dots.count, *painterLength, .......
        this.mainData = [[4,0,9.0000,18.0000,"Stepped"]];    // *mainData.count, *Dots.count, *painterLength, painterStepSize, .......
            this.mainData.push(["spline",[0.0000,4.5000],[0.0000,-4.5000]]);
            this.mainData.push(["spline",[0.0000,-4.5000],[9.0000,-4.5000]]);
            this.mainData.push(["spline",[9.0000,-4.5000],[9.0000,4.5000]]);
            this.mainData.push(["spline",[9.0000,4.5000],[0.0000,4.5000]]);
        this.endData = [[0,0,0.0000]];    // *endData.count, *Dots.count, *painterLength, .......
        this.endApexData = [[0,0,0.0000]];    // *endApexData.count, *Dots.count, *painterLength, .......
        this.knotData = [[0,0,0.0000,0,0.00]];    // *knotData.count, *Dots.count, *painterLength, insertions, insertPoints.......
        break;
    case "SqauredShapeFlexing":
        this.beginApexData = [[0,0,0.0000]];    // *beginApexData.count, *Dots.count, *painterLength, .......
        this.beginData = [[0,0,0.0000]];    // *beginData.count, *Dots.count, *painterLength, .......
        this.mainData = [[4,0,9.0000,18.0000,"Stepped"]];    // *mainData.count, *Dots.count, *painterLength, painterStepSize, .......
            this.mainData.push(["spline",[0.0000,4.5000],[0.0000,-4.5000]]);
            this.mainData.push(["spline",[0.0000,-4.5000],[1.5000,-4.5000],[3.0000,-4.5000],[4.5000,-4.5000],[6.0000,-4.5000],[7.5000,-4.5000],[9.0000,-4.5000]]);
            this.mainData.push(["spline",[9.0000,-4.5000],[9.0000,4.5000]]);
            this.mainData.push(["spline",[9.0000,4.5000],[7.5000,4.5000],[6.0000,4.5000],[4.5000,4.5000],[3.0000,4.5000],[1.5000,4.5000],[0.0000,4.5000]]);
        this.endData = [[0,0,0.0000]];    // *endData.count, *Dots.count, *painterLength, .......
        this.endApexData = [[0,0,0.0000]];    // *endApexData.count, *Dots.count, *painterLength, .......
        this.knotData = [[0,0,0.0000,0,0.00]];    // *knotData.count, *Dots.count, *painterLength, insertions, insertPoints.......
        break;
    case "TelephoneCoil":    // # ToDo #
        this.beginApexData = [[0,0,0.0000]];    // *beginApexData.count, *Dots.count, *painterLength, .......
        this.beginData = [[0,0,0.0000]];    // *beginData.count, *Dots.count, *painterLength, .......
        this.mainData = [[0,0,0.0000,0.0000,"Boxed"]];    // *mainData.count, *Dots.count, *painterLength, painterStepSize, .......
        this.endData = [[0,0,0.0000]];    // *endData.count, *Dots.count, *painterLength, .......
        this.endApexData = [[0,0,0.0000]];    // *endApexData.count, *Dots.count, *painterLength, .......
        this.knotData = [[0,0,0.0000,0,0.00]];    // *knotData.count, *Dots.count, *painterLength, insertions, insertPoints.......
        break;
    }    // End switch
// ** USER-ADJUSTABLE PARAMETERS ABOVE **
    return painterSet + " done";
};

/**
 * Adds a menu for this action to Examples/Math Examples/Painter.
 */
FlexPainter.init = function(basePath) {
    var action = new RGuiAction(qsTr("&FlexPainter"), RMainWindowQt.getMainWindow());
    action.setRequiresDocument(true);
    action.setScriptFile(basePath + "/FlexPainter.js");
    action.setIcon(basePath + "/FlexPainter.svg");
//    action.setDefaultShortcut(new QKeySequence("xyz, xyz"));    // # ToDo #
//    action.setDefaultCommands(["xyzxyz", "xyzxyz"]);    // # ToDo #
    action.setGroupSortOrder(79700);
    action.setSortOrder(300);
    action.setWidgetNames(["MathExamplesMenu"]);
};

FlexPainter.Remarks = function(Dummy) {
//    # Error 0 # MAJOR ISSUE 1 # Issue 18 # Solved 1+9 # Fixed 6 # ToDo 21 # Done 3 #
//    # Busy 6 # Reminder 1 # Workaround 5 #
// *************************************************************************************************
//    # MAJOR ISSUE # Dialog painter set selection coding and how to have scope...
//              JAVAscript/Qt related  Asked for help

//    # Issue # Can't find where qDebug("...") is sending to, can't find / don't know of a console nor std... app

//    # Issue # Two options
//              widgets["MainLengthX"].text = this.mainData[0][2] * this.mainScaleX;
//              widgets["MainLengthX"].setValue(this.mainData[0][2] * this.mainScaleX);
//              ... = widgets["FactorY"].text;
//              ... = widgets["FactorY"].getValue();

//    # Issue Fixed # Angle at polyline
//              shape.getAngleAtPoint(RVector) == NaN
//              shape.getAngleAt(distance,RS.FromStart) == NaN
//              # Workaround # Angle at segment

//    # Issue Fixed # castAngle failing @startPoint or @distance zero or @endpoint or @distance objlength
//              Some such castings return null vectors for castpoint and subsequently an error for castAngle
//              # ToDo Done # Isolate cases see FlexPainter.prototype.castPainter...switch(this.castingType){}
//              # ToDo # File bugreport =???= offset to poly @startPoint || @endpoint
//        Temporary workaround: AVOID the use of fitpoints on the reference margins
//        Final # Workaround # Offset tiny bit from start- or endpoint
//              # Issue # Have to offset >999 times RS.PointTolerance to stay away from first segment
//              Tested RS.PointTolerance times { 2,10,100,1000(ok),500,750,999,1000(ok) => 2000 times}
//              # ToDo # poly test with arcs, arc-line

//    # Issue Fixed # Non-grouped castings preffered to be displayed successively
//               Still true in debugger mode 
//               No longer in normal mode
//               # Workaround # with disableInput(), update(), enableInput()
//                   # Issue Fixed # Is slow
//                       # Workaround # omitted for only one and for last

//    # Issue Solved # metric vs imperial (mm stay mm or units stay units??? eg. fonts base 9)
//               Solved with: base 9 & Units
//               # ToDo Busy # Detect chages to former use v.37
//                             Could pose a problem !!!

//    # ToDo # Solution for Painter splines data sets.(Partially prepared UI & code)
//             Investigate splines in Blocks:
//                 "Splines" -> sequentially local fitpoints by splines
//                 "Dotsplines" -> points sorted in X are sequentially global fitpoints
//                 Others open to implement
//                 # Issue # As such no point or poly data possible

//    # ToDo # Shift placement 0-100% actual painter stepsize

//    # ToDo # Add painter entities types: Line, Point, Arc, Circle, Poly

//    # ToDo # Insertable painters (eg. Knots) @(aaa%,bbb%,ccc%,...)(Partially prepared UI & code)

//    # ToDo # With fitN rather fit N or N+1 instead of cast until end ??? pro - con

//    # ToDo Busy # Check user numerical entries ... mainScaleX/Y != -n; 0; huge v.37
//            # ToDo # Investigate '.setValidator' for QLineEdit (Occurs in eg. InvoluteSpur.js)
//            Done: Can't start entries with '-', can enter expression that gives negative result v.37
//            Done: Made absolute and factor limited to 0.01 & 111 and written back to Widget v.38
//            Done: limited casting main painter to 9999 / entity  v.38
//                  (On short line in test f0.01 14s 21139 splines /3 is about 7000 painters)
//            Done: Dirty loop untill flawless v.39

// *************************************************************************************************
//    # MAJOR ISSUE Solved # AppCrash on casting Null-splines followed by:
//                    - Select All
//                    - Select Layer Entities: cast layer
//                    - Invert Selection (with base entity(-ies) still selected)
//                    - Selection filter: entities on cast layer
//                    - Selection filter: splines
//                    - draw select including origin
//                    NOT draw select excluding origin
//              Solved with: Difference in target
//              # ToDo # Debate not catching NULL entry
//    # Issue Solved # Division Pi/2 was used 3 times in each painter spline fitpoint casting loop
//              Solved with: appropriate math, script global this.var, locals as var and some as const (faster)
//              Still interesting for overal speed
//              Exec time: factor x1: literal (1.570796...)
//                         factor +/-x2: const orthoAngle (not Global)
//                         factor +/-x4: var orthoAngle && const orthoAngle (Global) && this.orthoAngle
//                         factor +/-x8...12: Math.PI/2
//    # Issue Solved # 'if () {...}' block with multiple AND conditions take longer to evaluate.
//              Solved for AND in frequent looping with: Multistage testing, test larger variations first
//              Solved for OR with: Test more likely to be true first
//              Kept for Main script loop
//              # Reminder # Check 'if() {...}' block conditions
//    # Issue Solved # 'For(){...}' loops work faster with less calculations in the conditions
//              # Reminder # Check 'For() {...}' loop conditions
//    # Issue Solved # Pushing messages to command history is slow (80ms/entry vs casting fitpoints: 10/ms)
//               Solved with beVerbose flag. Super persistently set
//               If not critical or essential 
//    # ToDo Done # Optionally delete base entity(-ies) in one operation at end. Super persistently cleared
//    # ToDo Done # Beside dynamically scale Open Forms optional Align @start @end @center

// ***Reminders*************************************************************************************
//    !! Use '==' not '=' !! apart from the fact '==' is allways conditional
//    Don't use '>>' by mistake nor '<<' '=<' '=>'
//    Code on lines must end with ';' or '}' and nothing else unless the code is properly multilined
//    Simple test: even counts of '[', ']', '(', ')', '{', '}'   forget 1 ')' will lead to crash
//    'this.argument' in FlexPainter.prototype.xyz functions    Scope: all FlexPainter.prototype. functions
//    const only have function scope
//    bool true = multipler 1, bool false = multipler 0 BUT TAKE CARE
//    Return more arguments var ret = []; ret.push(...); ret.push(...); return ret;
//    RMath.fuzzyCompare(value, value, tolerance);  RS.PointTolerance
//    Math.round rounds correct towards nearest integer
//    Math.trunc simply cuts of the point and the digits. Is not done by Math!
//    Progress see Ucs.js ??? ISNOT part of EAction Class. See RProgressHandler Class 
//        EAction.setProgressText("Applying operation...");
//        EAction.setProgress((100 / ids.length) *  i);
//        EAction.setProgress(100);
//        EAction.setProgressEnd();

};